@@ -22,12 +22,12 @@ except ImportError:  # pragma: no cover
     anysqlite = None  # type: ignore
 
 from httpcore import Request, Response
-from typing_extensions import TypeAlias
 
-from hishel._serializers import BaseSerializer, clone_model
+if tp.TYPE_CHECKING:  # pragma: no cover
+    from typing_extensions import TypeAlias
 
 from .._files import AsyncFileManager
-from .._serializers import JSONSerializer, Metadata
+from .._serializers import BaseSerializer, JSONSerializer, Metadata, clone_model
 from .._synchronization import AsyncLock
 from .._utils import float_seconds_to_int_milliseconds
 
@@ -36,13 +36,14 @@ logger = logging.getLogger("hishel.storages")
 __all__ = (
     "AsyncBaseStorage",
     "AsyncFileStorage",
-    "AsyncRedisStorage",
-    "AsyncSQLiteStorage",
     "AsyncInMemoryStorage",
+    "AsyncRedisStorage",
     "AsyncS3Storage",
+    "AsyncSQLiteStorage",
 )
 
 StoredResponse: TypeAlias = tp.Tuple[Response, Request, Metadata]
+RemoveTypes = tp.Union[str, Response]
 
 try:
     import redis.asyncio as redis
@@ -62,6 +63,9 @@ class AsyncBaseStorage:
     async def store(self, key: str, response: Response, request: Request, metadata: Metadata | None = None) -> None:
         raise NotImplementedError()
 
+    async def remove(self, key: RemoveTypes) -> None:
+        raise NotImplementedError()
+
     async def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
         raise NotImplementedError()
 
@@ -99,8 +103,7 @@ class AsyncFileStorage(AsyncBaseStorage):
         self._base_path = Path(base_path) if base_path is not None else Path(".cache/hishel")
         self._gitignore_file = self._base_path / ".gitignore"
 
-        if not self._base_path.is_dir():
-            self._base_path.mkdir(parents=True)
+        self._base_path.mkdir(parents=True, exist_ok=True)
 
         if not self._gitignore_file.is_file():
             with open(self._gitignore_file, "w", encoding="utf-8") as f:
@@ -137,6 +140,23 @@ class AsyncFileStorage(AsyncBaseStorage):
             )
         await self._remove_expired_caches(response_path)
 
+    async def remove(self, key: RemoveTypes) -> None:
+        """
+        Removes the response from the cache.
+
+        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
+        :type key: Union[str, Response]
+        """
+
+        if isinstance(key, Response):  # pragma: no cover
+            key = tp.cast(str, key.extensions["cache_metadata"]["cache_key"])
+
+        response_path = self._base_path / key
+
+        async with self._lock:
+            if response_path.exists():
+                response_path.unlink(missing_ok=True)
+
     async def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
         """
         Updates the metadata of the stored response.
@@ -198,16 +218,20 @@ class AsyncFileStorage(AsyncBaseStorage):
             if response_path.is_file():
                 age = time.time() - response_path.stat().st_mtime
                 if age > self._ttl:
-                    response_path.unlink()
+                    response_path.unlink(missing_ok=True)
             return
 
         self._last_cleaned = time.monotonic()
         async with self._lock:
-            for file in self._base_path.iterdir():
-                if file.is_file():
-                    age = time.time() - file.stat().st_mtime
-                    if age > self._ttl:
-                        file.unlink()
+            with os.scandir(self._base_path) as entries:
+                for entry in entries:
+                    try:
+                        if entry.is_file():
+                            age = time.time() - entry.stat().st_mtime
+                            if age > self._ttl:
+                                os.unlink(entry.path)
+                    except FileNotFoundError:  # pragma: no cover
+                        pass
 
 
 class AsyncSQLiteStorage(AsyncBaseStorage):
@@ -225,16 +249,14 @@ class AsyncSQLiteStorage(AsyncBaseStorage):
     def __init__(
         self,
         serializer: tp.Optional[BaseSerializer] = None,
-        connection: tp.Optional["anysqlite.Connection"] = None,
+        connection: tp.Optional[anysqlite.Connection] = None,
         ttl: tp.Optional[tp.Union[int, float]] = None,
     ) -> None:
         if anysqlite is None:  # pragma: no cover
             raise RuntimeError(
-                (
-                    f"The `{type(self).__name__}` was used, but the required packages were not found. "
-                    "Check that you have `Hishel` installed with the `sqlite` extension as shown.\n"
-                    "```pip install hishel[sqlite]```"
-                )
+                f"The `{type(self).__name__}` was used, but the required packages were not found. "
+                "Check that you have `Hishel` installed with the `sqlite` extension as shown.\n"
+                "```pip install hishel[sqlite]```"
             )
         super().__init__(serializer, ttl)
 
@@ -249,7 +271,7 @@ class AsyncSQLiteStorage(AsyncBaseStorage):
                 if not self._connection:  # pragma: no cover
                     self._connection = await anysqlite.connect(".hishel.sqlite", check_same_thread=False)
                 await self._connection.execute(
-                    ("CREATE TABLE IF NOT EXISTS cache(key TEXT, data BLOB, date_created REAL)")
+                    "CREATE TABLE IF NOT EXISTS cache(key TEXT, data BLOB, date_created REAL)"
                 )
                 await self._connection.commit()
                 self._setup_completed = True
@@ -284,6 +306,24 @@ class AsyncSQLiteStorage(AsyncBaseStorage):
             await self._connection.commit()
         await self._remove_expired_caches()
 
+    async def remove(self, key: RemoveTypes) -> None:
+        """
+        Removes the response from the cache.
+
+        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
+        :type key: Union[str, Response]
+        """
+
+        await self._setup()
+        assert self._connection
+
+        if isinstance(key, Response):  # pragma: no cover
+            key = tp.cast(str, key.extensions["cache_metadata"]["cache_key"])
+
+        async with self._lock:
+            await self._connection.execute("DELETE FROM cache WHERE key = ?", [key])
+            await self._connection.commit()
+
     async def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
         """
         Updates the metadata of the stored response.
@@ -335,8 +375,8 @@ class AsyncSQLiteStorage(AsyncBaseStorage):
             return self._serializer.loads(cached_response)
 
     async def aclose(self) -> None:  # pragma: no cover
-        assert self._connection
-        await self._connection.close()
+        if self._connection is not None:
+            await self._connection.close()
 
     async def _remove_expired_caches(self) -> None:
         assert self._connection
@@ -363,16 +403,14 @@ class AsyncRedisStorage(AsyncBaseStorage):
     def __init__(
         self,
         serializer: tp.Optional[BaseSerializer] = None,
-        client: tp.Optional["redis.Redis"] = None,  # type: ignore
+        client: tp.Optional[redis.Redis] = None,  # type: ignore
         ttl: tp.Optional[tp.Union[int, float]] = None,
     ) -> None:
         if redis is None:  # pragma: no cover
             raise RuntimeError(
-                (
-                    f"The `{type(self).__name__}` was used, but the required packages were not found. "
-                    "Check that you have `Hishel` installed with the `redis` extension as shown.\n"
-                    "```pip install hishel[redis]```"
-                )
+                f"The `{type(self).__name__}` was used, but the required packages were not found. "
+                "Check that you have `Hishel` installed with the `redis` extension as shown.\n"
+                "```pip install hishel[redis]```"
             )
         super().__init__(serializer, ttl)
 
@@ -408,6 +446,19 @@ class AsyncRedisStorage(AsyncBaseStorage):
             key, self._serializer.dumps(response=response, request=request, metadata=metadata), px=px
         )
 
+    async def remove(self, key: RemoveTypes) -> None:
+        """
+        Removes the response from the cache.
+
+        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
+        :type key: Union[str, Response]
+        """
+
+        if isinstance(key, Response):  # pragma: no cover
+            key = tp.cast(str, key.extensions["cache_metadata"]["cache_key"])
+
+        await self._client.delete(key)
+
     async def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
         """
         Updates the metadata of the stored response.
@@ -424,7 +475,9 @@ class AsyncRedisStorage(AsyncBaseStorage):
 
         ttl_in_milliseconds = await self._client.pttl(key)
 
-        if ttl_in_milliseconds == -2:  # pragma: no cover
+        # -2: if the key does not exist in Redis
+        # -1: if the key exists in Redis but has no expiration
+        if ttl_in_milliseconds == -2 or ttl_in_milliseconds == -1:  # pragma: no cover
             await self.store(key, response, request, metadata)
         else:
             await self._client.set(
@@ -450,7 +503,7 @@ class AsyncRedisStorage(AsyncBaseStorage):
         return self._serializer.loads(cached_response)
 
     async def aclose(self) -> None:  # pragma: no cover
-        await self._client.close()
+        await self._client.aclose()
 
 
 class AsyncInMemoryStorage(AsyncBaseStorage):
@@ -506,6 +559,20 @@ class AsyncInMemoryStorage(AsyncBaseStorage):
             self._cache.put(key, (stored_response, time.monotonic()))
         await self._remove_expired_caches()
 
+    async def remove(self, key: RemoveTypes) -> None:
+        """
+        Removes the response from the cache.
+
+        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
+        :type key: Union[str, Response]
+        """
+
+        if isinstance(key, Response):  # pragma: no cover
+            key = tp.cast(str, key.extensions["cache_metadata"]["cache_key"])
+
+        async with self._lock:
+            self._cache.remove_key(key)
+
     async def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
         """
         Updates the metadata of the stored response.
@@ -583,6 +650,8 @@ class AsyncS3Storage(AsyncBaseStorage):  # pragma: no cover
     :type check_ttl_every: tp.Union[int, float]
     :param client: A client for S3, defaults to None
     :type client: tp.Optional[tp.Any], optional
+    :param path_prefix: A path prefix to use for S3 object keys, defaults to "hishel-"
+    :type path_prefix: str, optional
     """
 
     def __init__(
@@ -592,16 +661,15 @@ class AsyncS3Storage(AsyncBaseStorage):  # pragma: no cover
         ttl: tp.Optional[tp.Union[int, float]] = None,
         check_ttl_every: tp.Union[int, float] = 60,
         client: tp.Optional[tp.Any] = None,
+        path_prefix: str = "hishel-",
     ) -> None:
         super().__init__(serializer, ttl)
 
         if boto3 is None:  # pragma: no cover
             raise RuntimeError(
-                (
-                    f"The `{type(self).__name__}` was used, but the required packages were not found. "
-                    "Check that you have `Hishel` installed with the `s3` extension as shown.\n"
-                    "```pip install hishel[s3]```"
-                )
+                f"The `{type(self).__name__}` was used, but the required packages were not found. "
+                "Check that you have `Hishel` installed with the `s3` extension as shown.\n"
+                "```pip install hishel[s3]```"
             )
 
         self._bucket_name = bucket_name
@@ -611,6 +679,7 @@ class AsyncS3Storage(AsyncBaseStorage):  # pragma: no cover
             bucket_name=bucket_name,
             is_binary=self._serializer.is_binary,
             check_ttl_every=check_ttl_every,
+            path_prefix=path_prefix,
         )
         self._lock = AsyncLock()
 
@@ -638,6 +707,20 @@ class AsyncS3Storage(AsyncBaseStorage):  # pragma: no cover
 
         await self._remove_expired_caches(key)
 
+    async def remove(self, key: RemoveTypes) -> None:
+        """
+        Removes the response from the cache.
+
+        :param key: Hashed value of concatenated HTTP method and URI or an HTTP response
+        :type key: Union[str, Response]
+        """
+
+        if isinstance(key, Response):  # pragma: no cover
+            key = tp.cast(str, key.extensions["cache_metadata"]["cache_key"])
+
+        async with self._lock:
+            await self._s3_manager.remove_entry(key)
+
     async def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
         """
         Updates the metadata of the stored response.
